#include <string>
#include <list>
#include <vector>
#include <fstream>
#include <iostream>
#include "SimpleRecommender.h";
#include "RatingInformation.cc";
//#include "MovieInformation.h";
using namespace std;

class MovieComparer
{
    private : 
        int id;
    public : 
        MovieComparer(int init_id) { 
            id = init_id;
        }
        
        bool operator()(const RatingInformation& rating )
        {
            return rating.movieId != id;
        }
};

class RatingContainedChecker
{
    private : 
        list<RatingInformation> ratings;
    public : 
        RatingContainedChecker(list<RatingInformation> allRatings) { 
            ratings = allRatings;
        }
        
        bool operator()(const RatingInformation& rating )
        {
            //check if there is a rating in ratings with matching id/movie id
            for (list<RatingInformation>::iterator current = ratings.begin(); current != ratings.end(); current++)
            {
                if (current->movieId == rating.movieId && current->userId == rating.userId)
                {
                    return true;
                }
            }
            
            return false;
        }
};

bool comparePredictions(RatingInformation r1, RatingInformation r2)
{
    return r1.rating < r2.rating;
}

double SimpleRecommender::generatePrediction(int userid, string movieName)
{
    //find the movie with the right id.
    int id = getMovieId(movieName);
    cout << id << endl;
    
    //if the id is less than 0 then there is no such movie
    if (id < 0)
    {
        return -1;
    }
    
    //otherwise we will take the average of all the ratings for this movie.
    MovieComparer m(id);
    int count = ratings.size() - count_if(ratings.begin(), ratings.end(), m);
    vector<RatingInformation> moviesWithId(count);
    remove_copy_if(ratings.begin(), ratings.end(), moviesWithId.begin(), m);
 
   //now take the average of all of them    
    double sum = 0;
    for (vector<RatingInformation>::iterator current = moviesWithId.begin(); current != moviesWithId.end(); current++)
    {
        sum = sum + current->rating;
    }
    
    return sum / count;
}

vector<string> SimpleRecommender::makeRecommendations(int userid, int n)
{
    //should choose the top n recommendations. 
    //This means we should first calculate predictions for all movies, then sort, then choose a subset.
    //Then convert back to list<string>
    vector<RatingInformation> predictions;
    for (vector<MovieInformation>::const_iterator current = movies.begin(); current != movies.end(); current++)
    {
        RatingInformation rating;
        rating.userId = userid;
        rating.movieId = current->getId();
        rating.rating = generatePrediction(userid, current->getName());
        predictions.push_back(rating);
    }

    //now sort based on the predictions. First let's remove all elements that were already rated by the user.
    RatingContainedChecker ratingContainedChecker(ratings);
    vector<RatingInformation>::iterator newEnd = remove_if(predictions.begin(), predictions.end(), ratingContainedChecker);
    sort(predictions.begin(), newEnd, comparePredictions);
    predictions.resize(newEnd - predictions.begin());
    vector<string> moviePredictions;
    for (int i = 0; i < n; i++)
    {
        if (i >= predictions.size())
        {
            return moviePredictions;
        }
        
        string name = movies.at(predictions.at(i).movieId).getName();
        moviePredictions.push_back(name);
    }
    
    return moviePredictions;
}

int SimpleRecommender::getMovieId(string movieName)
{
    for (vector<MovieInformation>::iterator current = movies.begin(); current != movies.end(); current++)
    {
        if (current->getName() == movieName)
        {
            return current->getId();
        }
    }
    
    return -1;
}

void SimpleRecommender::loadMovieListFromFile(string file)
{
    ifstream in_file;
    in_file.open(file.c_str());
    movies.clear();
    string line;
    int i = 0;
    while (getline(in_file, line))
    {
        MovieInformation m(line,i);
        movies.push_back(m);
        i++;
    }
    
    in_file.close();
}

void SimpleRecommender::loadRatingsFromFile(string file)
{
    ratings.clear();
    ifstream in_file;
    in_file.open(file.c_str());
    
    while (!in_file.fail())
    {
        RatingInformation ratingInformation;
        in_file >> ratingInformation.userId;
        
        if (in_file.fail())
        {
            break;
        }
        
        in_file >> ratingInformation.movieId;
        in_file >> ratingInformation.rating;
        ratings.push_front(ratingInformation);
    }
    
    in_file.close();
}

void SimpleRecommender::printMovies()
{
    for (vector<MovieInformation>::iterator current = movies.begin(); current != movies.end(); current++)
    {
        cout << current->getName() << " " << current->getId() << endl;
    }
}

void SimpleRecommender::printRatings()
{
    for (list<RatingInformation>::iterator current = ratings.begin(); current != ratings.end(); current++)
    {
        cout << current->movieId << " " << current->userId << endl;
    }
}

/*int main()
{
    SimpleRecommender* sr = new SimpleRecommender();
    sr->loadMovieListFromFile("sample.txt");
    sr->printMovies();
    sr->loadRatingsFromFile("TestFile.txt");
    sr->printRatings();
    cout << sr->generatePrediction(0, "Indiana Jones");
    delete sr;
}*/
